# This file has been generated


build_system = 'AppVeyor'
build_id = '29670643'
build_number = '559'
build_version = '1.0.559'
job_id = 'vkbtowmalruayk2s'
job_url = 'https://ci.appveyor.com/project/KyleAltendorf/pm/build/job/vkbtowmalruayk2s'
