# This file was automatically generated by sub\epyqlib\epyqlib\cli\generateversion.py
__version_build_time__ = '2019-12-20T03:44:17.108793'

__version__ = '2019.12.5'
__sha__ = 'a8161404103220464cef0b4cd71c20a845edc46e'
__revision__ = 'a8161404103220464cef0b4cd71c20a845edc46e'
